@@ -48,6 +48,8 @@ class _Constants:
 		BaseDirectory.save_data_path("zeitgeist"))
 	DATABASE_FILE = os.environ.get("ZEITGEIST_DATABASE_PATH",
 		os.path.join(DATA_PATH, "activity.sqlite"))
+	DATABASE_FILE_BACKUP = os.environ.get("ZEITGEIST_DATABASE_BACKUP_PATH",
+		os.path.join(DATA_PATH, "activity.sqlite.bck"))
 	DEFAULT_LOG_PATH = os.path.join(BaseDirectory.xdg_cache_home,
 		"zeitgeist", "daemon.log")
 	
@@ -57,7 +59,7 @@ class _Constants:
 	
 	# Required version of DB schema
 	CORE_SCHEMA="core"
-	CORE_SCHEMA_VERSION = 3
+	CORE_SCHEMA_VERSION = 4
 	
 	USER_EXTENSION_PATH = os.path.join(DATA_PATH, "extensions")
 	
