@@ -7,7 +7,7 @@
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
+# the Free Software Foundation, either version 2.1 of the License, or
 # (at your option) any later version.
 #
 # This program is distributed in the hope that it will be useful,
@@ -22,7 +22,6 @@ import os
 import logging
 import weakref # avoid circular references as they confuse garbage collection
 
-logging.basicConfig(level=logging.DEBUG)
 log = logging.getLogger("zeitgeist.extension")
 
 import zeitgeist
@@ -33,6 +32,19 @@ def safe_issubclass(obj, cls):
 		return issubclass(obj, cls)
 	except TypeError:
 		return False
+		
+def get_extension_name(extension):
+	""" We are using the name of the Extension-class as extension's unique
+	name, later we might want to prepend modul or package names.
+	"""
+	if safe_issubclass(extension, Extension):
+		return extension.__name__
+	elif isinstance(extension, Extension):
+		return extension.__class__.__name__
+	else:
+		raise TypeError(
+			"Can't identify %r, an extension has to be a subclass or "
+			"instance of extension.Extension" %extension)
 
 class Extension(object):
 	""" Base class for all extensions
@@ -45,11 +57,20 @@ class Extension(object):
 	inserted and retrieved from the log. These hooks can either block the
 	event completely, modify it, or add additional metadata to it.
 	"""
-	PUBLIC_METHODS = None
+	PUBLIC_METHODS = []
 	
 	def __init__(self, engine):
 		self.engine = weakref.proxy(engine)
 	
+	def unload(self):
+		"""
+		This method gets called before Zeitgeist stops.
+		
+		Execution of this method isn't guaranteed, and you shouldn't do
+		anything slow in there.
+		"""
+		pass
+	
 	def pre_insert_event(self, event, sender):
 		"""
 		Hook applied to all events before they are inserted into the
@@ -162,21 +183,21 @@ def _scan_extensions():
 	
 	# Find system extensions
 	log.debug("Searching for system extensions in: %s" % config.extensiondir)
-	sys_modules = filter(lambda m : m.endswith(".py"), os.listdir(config.extensiondir))
-	sys_modules = modules = map(lambda m : "_zeitgeist.engine.extensions." + m.rpartition(".")[0], sys_modules)
+	sys_modules = filter(lambda m: m.endswith(".py"), os.listdir(config.extensiondir))
+	sys_modules = map(lambda m: "_zeitgeist.engine.extensions." + m.rpartition(".")[0], sys_modules)
 	
 	# Find user extensions
 	log.debug("Searching for user extensions in: %s" % constants.USER_EXTENSION_PATH)
 	user_modules = []
 	try:
-		user_modules = filter(lambda m : m.endswith(".py"), os.listdir(os.path.expanduser(constants.USER_EXTENSION_PATH)))
-		user_modules = map(lambda m : m.rpartition(".")[0], user_modules)
+		user_modules = filter(lambda m: m.endswith(".py"), os.listdir(os.path.expanduser(constants.USER_EXTENSION_PATH)))
+		user_modules = map(lambda m: m.rpartition(".")[0], user_modules)
 	except OSError:
 		pass # USER_EXTENSION_PATH doesn't exist
 	
 	# If we have module conflicts let the user extensions win,
 	# and remove the system provided extension from our list
-	user_module_names = map(lambda m : os.path.basename(m), user_modules)
+	user_module_names = map(lambda m: os.path.basename(m), user_modules)
 	for mod in list(sys_modules):
 		mod_name = mod.rpartition(".")[2]
 		if mod_name in user_module_names:
@@ -197,7 +218,6 @@ def _scan_extensions():
 			if safe_issubclass(cls, Extension) and not cls is Extension:
 				extensions.append(cls)
 	return extensions
-			
 
 def _load_class(path):
 	"""
@@ -236,22 +256,20 @@ class ExtensionsCollection(object):
 		return "%s(%r)" %(self.__class__.__name__, sorted(self.__methods.keys()))
 			
 	def load(self, extension):
-		log.debug("Loading extension '%s'" % extension.__name__)
 		if not issubclass(extension, Extension):
 			raise TypeError("Unable to load %r, all extensions must be "
 				"subclasses of %r" % (extension, Extension))
-		if getattr(extension, "PUBLIC_METHODS", None) is None:
-			raise ValueError("Unable to load %r, this extension has not "
-				"defined any methods" % extension)
+		ext_name = get_extension_name(extension)
+		log.debug("Loading extension '%s'" %ext_name)
 		try:
 			obj = extension(self.__engine)
 		except Exception:
-			log.exception("Failed loading the '%s' extension" % extension.__name__)
+			log.exception("Failed loading the '%s' extension" %ext_name)
 			return False
 
 		for method in obj.PUBLIC_METHODS:
 			self._register_method(method, getattr(obj, method))
-		self.__extensions[obj.__class__.__name__] = obj
+		self.__extensions[ext_name] = obj
 		
 	def unload(self, extension=None):
 		"""
@@ -265,18 +283,13 @@ class ExtensionsCollection(object):
 			
 			# We need to clone the key list to avoid concurrent
 			# modification of the extension dict
-			for ext_name in list(self.__extensions.iterkeys()):
-				self.unload(self.__extensions[ext_name])
+			for ext in list(self):
+				self.unload(ext)
 		else:
-			log.debug("Unloading extension '%s'" \
-					  % extension.__class__.__name__)
-			if safe_issubclass(extension, Extension):
-				ext_name = extension.__name__
-			elif isinstance(extension, Extension):
-				ext_name = extension.__class__.__name__
-			else:
-				raise TypeError
+			ext_name = get_extension_name(extension)
+			log.debug("Unloading extension '%s'" % ext_name)
 			obj = self.__extensions[ext_name]
+			obj.unload()
 			for method in obj.PUBLIC_METHODS:
 				del self.methods[method]
 			del self.__extensions[ext_name]
@@ -348,3 +361,6 @@ class ExtensionsCollection(object):
 	
 	def __iter__(self):
 		return self.__extensions.itervalues()
+		
+	def iter_names(self):
+		return self.__extensions.iterkeys()
