@@ -23,6 +23,7 @@ import os
 import cPickle as pickle
 import dbus
 import dbus.service
+import gobject
 import logging
 
 from zeitgeist.datamodel import get_timestamp_for_now
@@ -36,6 +37,7 @@ DATA_FILE = os.path.join(constants.DATA_PATH, "datasources.pickle")
 REGISTRY_DBUS_OBJECT_PATH = "/org/gnome/zeitgeist/data_source_registry"
 REGISTRY_DBUS_INTERFACE = "org.gnome.zeitgeist.DataSourceRegistry"
 SIG_FULL_DATASOURCE = "(sssa("+constants.SIG_EVENT+")bxb)"
+DISK_WRITE_TIMEOUT = 5*60*1000 # 5 Minutes
 
 class DataSource(OrigDataSource):
 	@classmethod
@@ -64,7 +66,7 @@ class DataSourceRegistry(Extension, dbus.service.Object):
 	:const:`org.gnome.zeitgeist.DataSourceRegistry`.
 	"""
 	PUBLIC_METHODS = ["register_data_source", "get_data_sources",
-		"set_data_source_enabled"]
+		"set_data_source_enabled", "get_data_source_from_id"]
 	
 	def __init__ (self, engine):
 	
@@ -88,16 +90,25 @@ class DataSourceRegistry(Extension, dbus.service.Object):
 		
 		# Connect to client disconnection signals
 		dbus.SessionBus().add_signal_receiver(self._name_owner_changed,
-		    signal_name="NameOwnerChanged",
-		    dbus_interface=dbus.BUS_DAEMON_IFACE,
-		    arg2="", # only match services with no new owner
-	    )
-	
+			signal_name="NameOwnerChanged",
+			dbus_interface=dbus.BUS_DAEMON_IFACE,
+			arg2="", # only match services with no new owner
+		)
+		
+		self._dirty = True
+		gobject.timeout_add(DISK_WRITE_TIMEOUT, self._write_to_disk)
+		
 	def _write_to_disk(self):
 		data = [DataSource.get_plain(datasource) for datasource in
 			self._registry.itervalues()]
-		with open(DATA_FILE, "w") as data_file:
-			pickle.dump(data, data_file)
+		try:
+			if self._dirty:
+				self._dirty = False
+				with open(DATA_FILE, "w") as data_file:
+					pickle.dump(data, data_file)
+		except Exception, e:
+			log.warn("Failed to write to data file %s: %s" % (DATA_FILE, e))
+		return True
 		#log.debug("Data-source registry update written to disk.")
 	
 	def pre_insert_event(self, event, sender):
@@ -106,6 +117,7 @@ class DataSourceRegistry(Extension, dbus.service.Object):
 				datasource = self._registry[unique_id]
 				# Update LastSeen time
 				datasource.last_seen = get_timestamp_for_now()
+				self._dirty = True
 				# Check whether the data-source is allowed to insert events
 				if not datasource.enabled:
 					return None
@@ -113,7 +125,7 @@ class DataSourceRegistry(Extension, dbus.service.Object):
 	
 	def unload(self):
 		self._write_to_disk()
-	
+
 	# PUBLIC
 	def register_data_source(self, unique_id, name, description, templates):
 		source = DataSource(str(unique_id), unicode(name), unicode(description),
@@ -140,7 +152,14 @@ class DataSourceRegistry(Extension, dbus.service.Object):
 			datasource.enabled = enabled
 			self.DataSourceEnabled(datasource.unique_id, enabled)
 		return True
-	
+
+	# PUBLIC
+	def get_data_source_from_id(self, unique_id):
+		if unique_id in self._registry:
+			return self._registry[unique_id]
+		else:
+			raise KeyError("DataSource not found: %s" % unique_id)
+
 	@dbus.service.method(REGISTRY_DBUS_INTERFACE,
 						 in_signature="sssa("+constants.SIG_EVENT+")",
 						 out_signature="b",
@@ -199,6 +218,22 @@ class DataSourceRegistry(Extension, dbus.service.Object):
 		"""
 		self.set_data_source_enabled(unique_id, enabled)
 
+	@dbus.service.method(REGISTRY_DBUS_INTERFACE,
+						 in_signature="s",
+						 out_signature=SIG_FULL_DATASOURCE,
+						 sender_keyword="sender")
+	def GetDataSourceFromId(self, unique_id, sender):
+		"""
+		Get a specific data-source from its unique ID.
+		If no data-source has been registered with specified unique
+		ID, a KeyError is raised.
+
+		:param unique_id: unique string identifying a data-source
+		:type unique_id: string
+		:returns: class:`DataSource <zeitgeist.datamodel.DataSource>` object
+		"""
+		return self.get_data_source_from_id(unique_id)
+
 	@dbus.service.signal(REGISTRY_DBUS_INTERFACE,
 						signature="sb")
 	def DataSourceEnabled(self, value, enabled):
