@@ -5,7 +5,10 @@
 # Copyright © 2009 Mikkel Kamstrup Erlandsen <mikkel.kamstrup@gmail.com>
 # Copyright © 2009-2010 Siegfried-Angel Gevatter Pujals <rainct@ubuntu.com>
 # Copyright © 2009-2011 Seif Lotfy <seif@lotfy.com>
-# Copyright © 2009 Markus Korn <thekorn@gmx.net>
+# Copyright © 2009-2011 Markus Korn <thekorn@gmx.net>
+# Copyright © 2011 Collabora Ltd.
+#             By Seif Lotfy <seif@lotfy.com>
+#             By Siegfried-Angel Gevatter Pujals <rainct@ubuntu.com>
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
@@ -26,6 +29,7 @@ import sys
 import os
 import logging
 from collections import defaultdict
+from array import array
 
 from zeitgeist.datamodel import Event as OrigEvent, StorageState, TimeRange, \
 	ResultType, get_timestamp_for_now, Interpretation, Symbol, NEGATION_OPERATOR, WILDCARD
@@ -80,17 +84,11 @@ def parse_operators(kind, field, value):
 	on query values, and handles the special case of Subject.Text correctly.
 	returns a (value_without_negation_and_wildcard, negation, wildcard)-tuple
 	"""
-	try:
-		value, negation = parse_negation(kind, field, value)
-	except ValueError:
-		if kind is Subject and field == Subject.Text:
-			# we do not support negation of the text field,
-			# the text field starts with the NEGATION_OPERATOR
-			# so we handle this string as the content instead
-			# of an operator
-			negation = False
-		else:
-			raise
+	if kind is Subject and field == Subject.Text:
+		# Negation and wildcard characters in the subject text are
+		# handled as plain text.
+		return value, False, False
+	value, negation = parse_negation(kind, field, value)
 	value, wildcard = parse_wildcard(kind, field, value)
 	return value, negation, wildcard
 
@@ -113,24 +111,32 @@ class ZeitgeistEngine:
 			# to be 100% sure.
 			raise RuntimeError("old database version")
 		
-		# Load extensions
-		default_extensions = get_extensions()
-		self.__extensions = ExtensionsCollection(self,
-		                                         defaults=default_extensions)
-		
 		self._interpretation = TableLookup(cursor, "interpretation")
 		self._manifestation = TableLookup(cursor, "manifestation")
 		self._mimetype = TableLookup(cursor, "mimetype")
 		self._actor = TableLookup(cursor, "actor")
 		self._event_cache = LRUCache(constants.CACHE_SIZE)
+
+		# Load extensions
+		default_extensions = get_extensions()
+		self.__extensions = ExtensionsCollection(self,
+			defaults=default_extensions)
 	
 	@property
 	def extensions(self):
 		return self.__extensions
 	
 	def close(self):
-		log.debug("Closing down the zeitgeist engine...")
+		if self._cursor is None:
+			log.debug("The Zeitgeist engine is closed already")
+			return
+		log.debug("Closing down the Zeitgeist engine...")
 		self.extensions.unload()
+		# make sure all symbol table are in good shape
+		# this Exception should never be raised, it would indicate a bug
+		# in the workaround for (LP: #598666)
+		if self._cursor.execute("SELECT * FROM _fix_cache").fetchone():
+			raise RuntimeError("Symbol cache is in bad state")
 		self._cursor.connection.close()
 		self._cursor = None
 		unset_cursor()
@@ -156,6 +162,7 @@ class ZeitgeistEngine:
 				log.error("Event %i broken: Table %s has no id %i" \
 						%(row["id"], field, row[field]))
 				return None
+		event.origin = row["event_origin_uri"] or ""
 		event.payload = row["payload"] or "" # default payload: empty string
 		return event
 	
@@ -163,7 +170,9 @@ class ZeitgeistEngine:
 		subject = Subject()
 		for field in ("uri", "text", "storage"):
 			setattr(subject, field, row["subj_" + field])
-		setattr(subject, "origin", row["subj_origin_uri"])
+		subject.origin = row["subj_origin_uri"]
+		if row["subj_current_uri"]:
+			subject.current_uri = row["subj_current_uri"]
 		for field in ("interpretation", "manifestation", "mimetype"):
 			# Try to get subject attributes from row using the attributed field id
 			# If attribute does not exist we break the attribute fetching and return
@@ -188,8 +197,8 @@ class ZeitgeistEngine:
 			return []
 		
 		# Split ids into cached and uncached
-		uncached_ids = []
-		cached_ids = []
+		uncached_ids = array("i")
+		cached_ids = array("i")
 		
 		# If ids batch greater than MAX_CACHE_BATCH_SIZE ids ignore cache
 		use_cache = True
@@ -204,7 +213,7 @@ class ZeitgeistEngine:
 				else:
 					uncached_ids.append(id)
 		
-		id_hash = defaultdict(list)
+		id_hash = defaultdict(lambda: array("i"))
 		for n, id in enumerate(ids):
 			# the same id can be at multible places (LP: #673916)
 			# cache all of them
@@ -230,16 +239,18 @@ class ZeitgeistEngine:
 		rows = self._cursor.execute("""
 			SELECT * FROM event_view
 			WHERE id IN (%s)
-			""" % ",".join("%d" % id for id in uncached_ids)).fetchall()
+			""" % ",".join("%d" % _id for _id in uncached_ids))
 		
-		log.debug("Got %d raw events in %fs" % (len(rows), time.time()-t))
+		time_get_uncached = time.time() - t
 		t = time.time()
 		
 		t_get_event = 0
 		t_get_subject = 0
 		t_apply_get_hooks = 0
 		
+		row_counter = 0
 		for row in rows:
+			row_counter += 1
 			# Assumption: all rows of a same event for its different
 			# subjects are in consecutive order.
 			t_get_event -= time.time()
@@ -275,6 +286,7 @@ class ZeitgeistEngine:
 					# at a decent level
 					
 
+		log.debug("Got %d raw events in %fs" % (row_counter, time_get_uncached))
 		log.debug("Got %d events in %fs" % (len(sorted_events), time.time()-t))
 		log.debug("    Where time spent in _get_event_from_row in %fs" % (t_get_event))
 		log.debug("    Where time spent in _get_subject_from_row in %fs" % (t_get_subject))
@@ -329,6 +341,10 @@ class ZeitgeistEngine:
 				if value:
 					subwhere.add_text_condition("actor", value, wildcard, negation, cache=self._actor)
 				
+				value, negation, wildcard = parse_operators(Event, Event.Origin, event_template.origin)
+				if value:
+					subwhere.add_text_condition("origin", value, wildcard, negation)
+				
 				if subject_templates is not None:
 					for subject_template in subject_templates:
 						value, negation, wildcard = parse_operators(Subject, Subject.Interpretation, subject_template.interpretation)
@@ -351,9 +367,6 @@ class ZeitgeistEngine:
 						if su_manif_where:
 							subwhere.extend(su_manif_where)
 						
-						# FIXME: Expand mime children as well.
-						# Right now we only do exact matching for mimetypes
-						# thekorn: this will be fixed when wildcards are supported
 						value, negation, wildcard = parse_operators(Subject, Subject.Mimetype, subject_template.mimetype)
 						if value:
 							subwhere.add_text_condition("subj_mimetype",
@@ -363,7 +376,12 @@ class ZeitgeistEngine:
 							value = getattr(subject_template, key)
 							if value:
 								value, negation, wildcard = parse_operators(Subject, getattr(Subject, key.title()), value)
-								subwhere.add_text_condition("subj_%s" %key, value, wildcard, negation)
+								subwhere.add_text_condition("subj_%s" % key, value, wildcard, negation)
+						
+						if subject_template.current_uri:
+							value, negation, wildcard = parse_operators(Subject,
+								Subject.CurrentUri, subject_template.current_uri)
+							subwhere.add_text_condition("subj_current_uri", value, wildcard, negation)
 						
 						if subject_template.storage:
 							subwhere.add_text_condition("subj_storage", subject_template.storage)
@@ -378,22 +396,25 @@ class ZeitgeistEngine:
 	
 	def _build_sql_event_filter(self, time_range, templates, storage_state):
 		
-		# FIXME: We need to take storage_state into account
-		if storage_state != StorageState.Any:
-			raise NotImplementedError
+		where = WhereClause(WhereClause.AND)
 		
 		# thekorn: we are using the unary operator here to tell sql to not use
 		# the index on the timestamp column at the first place. This `fix` for
 		# (LP: #672965) is based on some benchmarks, which suggest a performance
 		# win, but we might not oversee all implications.
 		# (see http://www.sqlite.org/optoverview.html section 6.0)
-		where = WhereClause(WhereClause.AND)
 		min_time, max_time = time_range
 		if min_time != 0:
 			where.add("+timestamp >= ?", min_time)
 		if max_time != sys.maxint:
 			where.add("+timestamp <= ?", max_time)
 		
+		if storage_state in (StorageState.Available, StorageState.NotAvailable):
+			where.add("(subj_storage_state = ? OR subj_storage_state IS NULL)",
+				storage_state)
+		elif storage_state != StorageState.Any:
+			raise ValueError, "Unknown storage state '%d'" % storage_state
+		
 		where.extend(self._build_sql_from_event_templates(templates))
 		
 		return where
@@ -424,54 +445,127 @@ class ZeitgeistEngine:
 		else:
 			raise NotImplementedError, "Unsupported return_mode."
 		
-		if order == ResultType.OldestActor:
-			sql += """
-				NATURAL JOIN (
-					SELECT actor, min(timestamp) AS timestamp
-					FROM event_view %s
-					GROUP BY actor)
-				""" % ("WHERE " + where.sql if where.sql else "")
-		elif where:
-			sql += " WHERE " + where.sql
+		wheresql = " WHERE %s" % where.sql if where else ""
 		
-		sql += (" ORDER BY timestamp DESC",
-			" ORDER BY timestamp ASC",
-			# thekorn: please note, event.subj_id == uri.id, as in
-			# the subj_id points directly to an entry in the uri table,
-			# so we are in fact grouping by subj_uris here
-			" GROUP BY subj_id ORDER BY timestamp DESC",
-			" GROUP BY subj_id ORDER BY timestamp ASC",
-			" GROUP BY subj_id ORDER BY COUNT(subj_id) DESC, timestamp DESC",
-			" GROUP BY subj_id ORDER BY COUNT(subj_id) ASC, timestamp ASC",
-			" GROUP BY actor ORDER BY COUNT(actor) DESC, timestamp DESC", 
-			" GROUP BY actor ORDER BY COUNT(actor) ASC, timestamp ASC",
-			" GROUP BY actor ORDER BY timestamp DESC",
-			" GROUP BY actor ORDER BY timestamp ASC",
-			" GROUP BY subj_origin ORDER BY timestamp DESC",
-			" GROUP BY subj_origin ORDER BY timestamp ASC",
-			" GROUP BY subj_origin ORDER BY COUNT(subj_origin) DESC, timestamp DESC",
-			" GROUP BY subj_origin ORDER BY COUNT(subj_origin) ASC, timestamp ASC",
-			" GROUP BY actor ORDER BY timestamp ASC",
-			" GROUP BY subj_interpretation ORDER BY timestamp DESC",
-			" GROUP BY subj_interpretation ORDER BY timestamp ASC",
-			" GROUP BY subj_interpretation ORDER BY COUNT(subj_interpretation) DESC",
-			" GROUP BY subj_interpretation ORDER BY COUNT(subj_interpretation) ASC",
-			" GROUP BY subj_mimetype ORDER BY timestamp DESC",
-			" GROUP BY subj_mimetype ORDER BY timestamp ASC",
-			" GROUP BY subj_mimetype ORDER BY COUNT(subj_mimetype) DESC",
-			" GROUP BY subj_mimetype ORDER BY COUNT(subj_mimetype) ASC")[order]
+		def group_and_sort(field, wheresql, time_asc=False, count_asc=None,
+			aggregation_type='max'):
+			
+			args = {
+				'field': field,
+				'aggregation_type': aggregation_type,
+				'where_sql': wheresql,
+				'time_sorting': 'ASC' if time_asc else 'DESC',
+				'aggregation_sql': '',
+				'order_sql': '',
+			}
+			
+			if count_asc is not None:
+				args['aggregation_sql'] = ', COUNT(%s) AS num_events' % \
+					field
+				args['order_sql'] = 'num_events %s,' % \
+					('ASC' if count_asc else 'DESC')
+			
+			return """
+				NATURAL JOIN (
+					SELECT %(field)s,
+						%(aggregation_type)s(timestamp) AS timestamp
+						%(aggregation_sql)s
+					FROM event_view %(where_sql)s
+					GROUP BY %(field)s)
+				GROUP BY %(field)s
+				ORDER BY %(order_sql)s timestamp %(time_sorting)s
+				""" % args
+		
+		if order == ResultType.MostRecentEvents:
+			sql += wheresql + " ORDER BY timestamp DESC"
+		elif order == ResultType.LeastRecentEvents:
+			sql += wheresql + " ORDER BY timestamp ASC"
+		elif order == ResultType.MostRecentEventOrigin:
+			sql += group_and_sort("origin", wheresql, time_asc=False)
+		elif order == ResultType.LeastRecentEventOrigin:
+			sql += group_and_sort("origin", wheresql, time_asc=True)
+		elif order == ResultType.MostPopularEventOrigin:
+			sql += group_and_sort("origin", wheresql, time_asc=False,
+				count_asc=False)
+		elif order == ResultType.LeastPopularEventOrigin:
+			sql += group_and_sort("origin", wheresql, time_asc=True,
+				count_asc=True)
+		elif order == ResultType.MostRecentSubjects:
+			# Remember, event.subj_id identifies the subject URI
+			sql += group_and_sort("subj_id", wheresql, time_asc=False)
+		elif order == ResultType.LeastRecentSubjects:
+			sql += group_and_sort("subj_id", wheresql, time_asc=True)
+		elif order == ResultType.MostPopularSubjects:
+			sql += group_and_sort("subj_id", wheresql, time_asc=False,
+				count_asc=False)
+		elif order == ResultType.LeastPopularSubjects:
+			sql += group_and_sort("subj_id", wheresql, time_asc=True,
+				count_asc=True)
+		elif order == ResultType.MostRecentCurrentUri:
+			sql += group_and_sort("subj_id_current", wheresql, time_asc=False)
+		elif order == ResultType.LeastRecentCurrentUri:
+			sql += group_and_sort("subj_id_current", wheresql, time_asc=True)
+		elif order == ResultType.MostPopularCurrentUri:
+			sql += group_and_sort("subj_id_current", wheresql, time_asc=False,
+				count_asc=False)
+		elif order == ResultType.LeastPopularCurrentUri:
+			sql += group_and_sort("subj_id_current", wheresql, time_asc=True,
+				count_asc=True)
+		elif order == ResultType.MostRecentActor:
+			sql += group_and_sort("actor", wheresql, time_asc=False)
+		elif order == ResultType.LeastRecentActor:
+			sql += group_and_sort("actor", wheresql, time_asc=True)
+		elif order == ResultType.MostPopularActor:
+			sql += group_and_sort("actor", wheresql, time_asc=False,
+				count_asc=False)
+		elif order == ResultType.LeastPopularActor:
+			sql += group_and_sort("actor", wheresql, time_asc=True,
+				count_asc=True)
+		elif order == ResultType.OldestActor:
+			sql += group_and_sort("actor", wheresql, time_asc=True,
+				aggregation_type="min")
+		elif order == ResultType.MostRecentOrigin:
+			sql += group_and_sort("subj_origin", wheresql, time_asc=False)
+		elif order == ResultType.LeastRecentOrigin:
+			sql += group_and_sort("subj_origin", wheresql, time_asc=True)
+		elif order == ResultType.MostPopularOrigin:
+			sql += group_and_sort("subj_origin", wheresql, time_asc=False,
+				count_asc=False)
+		elif order == ResultType.LeastPopularOrigin:
+			sql += group_and_sort("subj_origin", wheresql, time_asc=True,
+				count_asc=True)
+		elif order == ResultType.MostRecentSubjectInterpretation:
+			sql += group_and_sort("subj_interpretation", wheresql,
+				time_asc=False)
+		elif order == ResultType.LeastRecentSubjectInterpretation:
+			sql += group_and_sort("subj_interpretation", wheresql,
+				time_asc=True)
+		elif order == ResultType.MostPopularSubjectInterpretation:
+			sql += group_and_sort("subj_interpretation", wheresql,
+				time_asc=False, count_asc=False)
+		elif order == ResultType.LeastPopularSubjectInterpretation:
+			sql += group_and_sort("subj_interpretation", wheresql,
+				time_asc=True, count_asc=True)
+		elif order == ResultType.MostRecentMimeType:
+			sql += group_and_sort("subj_mimetype", wheresql, time_asc=False)
+		elif order == ResultType.LeastRecentMimeType:
+			sql += group_and_sort("subj_mimetype", wheresql, time_asc=True)
+		elif order == ResultType.MostPopularMimeType:
+			sql += group_and_sort("subj_mimetype", wheresql, time_asc=False,
+				count_asc=False)
+		elif order == ResultType.LeastPopularMimeType:
+			sql += group_and_sort("subj_mimetype", wheresql, time_asc=True,
+				count_asc=True)
 		
 		if max_events > 0:
 			sql += " LIMIT %d" % max_events
-		
-		result = self._cursor.execute(sql, where.arguments).fetchall()
+		result = array("i", self._cursor.execute(sql, where.arguments).fetch(0))
 		
 		if return_mode == 0:
 			log.debug("Found %d event IDs in %fs" % (len(result), time.time()- t))
-			result = [row[0] for row in result]
 		elif return_mode == 1:
-			log.debug("Found %d events IDs in %fs" % (len(result), time.time()- t))
-			result = self.get_events(ids=[row[0] for row in result], sender=sender)	
+			log.debug("Found %d events in %fs" % (len(result), time.time()- t))
+			result = self.get_events(ids=result, sender=sender)	
 		else:
 			raise Exception("%d" % return_mode)
 		
@@ -573,9 +667,23 @@ class ZeitgeistEngine:
 			raise ValueError("Illegal event format: No subject")
 		if not event.timestamp:
 			event.timestamp = get_timestamp_for_now()
-		
-		id = self.next_event_id()
-		event[0][Event.Id] = id		
+		if not event.interpretation == Interpretation.MOVE_EVENT:
+			for subject in event.subjects:
+				if not subject.current_uri:
+					subject.current_uri = subject.uri
+				elif not subject.uri == subject.current_uri:
+					raise ValueError("Illegal event: unless "
+						"event.interpretation is 'MOVE_EVENT' then subject.uri "
+						"and subject.current_uri have to be the same")
+		if event.interpretation == Interpretation.MOVE_EVENT:
+			for subject in event.subjects:
+				if subject.uri == subject.current_uri or not subject.current_uri:
+					raise ValueError("Redundant event: event.interpretation "
+						"indicates the uri has been moved yet the subject.uri "
+						"and subject.current_uri are identical")
+			
+		event_id = self.next_event_id()
+		event[0][Event.Id] = event_id
 		event = self.extensions.apply_pre_insert(event, sender)
 		if event is None:
 			raise AssertionError("Inserting of event was blocked by an extension")
@@ -586,9 +694,14 @@ class ZeitgeistEngine:
 		
 		# Make sure all URIs are inserted
 		_origin = [subject.origin for subject in event.subjects if subject.origin]
+		if event.origin:
+			_origin.append(event.origin)
+		_current_uri = [subject.current_uri
+			for subject in event.subjects if subject.current_uri]
 		self._cursor.execute("INSERT OR IGNORE INTO uri (value) %s"
-			% " UNION ".join(["SELECT ?"] * (len(event.subjects) + len(_origin))),
-			[subject.uri for subject in event.subjects] + _origin)
+			% " UNION ".join(["SELECT ?"] * (len(event.subjects) +
+				len(_origin) + len(_current_uri))), _origin +
+			[subject.uri for subject in event.subjects] + _current_uri)
 		
 		# Make sure all mimetypes are inserted
 		_mimetype = [subject.mimetype for subject in event.subjects \
@@ -610,15 +723,18 @@ class ZeitgeistEngine:
 				% " UNION ".join(["SELECT ?"] * len(_storage)), _storage)
 		
 		try:
-			for subject in event.subjects:	
+			for subject in event.subjects:
 				self._cursor.execute("""
 					INSERT INTO event (
 						id, timestamp, interpretation, manifestation, actor,
-						payload, subj_id,
+						origin, payload, subj_id, subj_id_current,
 						subj_interpretation, subj_manifestation, subj_origin,
 						subj_mimetype, subj_text, subj_storage
 					) VALUES (
-						?, ?, ?, ?, ?, ?,
+						?, ?, ?, ?, ?,
+						(SELECT id FROM uri WHERE value=?),
+						?,
+						(SELECT id FROM uri WHERE value=?),
 						(SELECT id FROM uri WHERE value=?),
 						?, ?,
 						(SELECT id FROM uri WHERE value=?),
@@ -626,22 +742,24 @@ class ZeitgeistEngine:
 						(SELECT id FROM text WHERE value=?),
 						(SELECT id from storage WHERE value=?)
 					)""", (
-						id,
+						event_id,
 						event.timestamp,
 						self._interpretation[event.interpretation],
 						self._manifestation[event.manifestation],
 						self._actor[event.actor],
+						event.origin,
 						payload_id,
 						subject.uri,
+						subject.current_uri,
 						self._interpretation[subject.interpretation],
 						self._manifestation[subject.manifestation],
 						subject.origin,
 						self._mimetype[subject.mimetype],
 						subject.text,
 						subject.storage))
-				
+			
 			self.extensions.apply_post_insert(event, sender)
-				
+			
 		except sqlite3.IntegrityError:
 			# The event was already registered.
 			# Rollback _last_event_id and return the ID of the original event
@@ -656,7 +774,25 @@ class ZeitgeistEngine:
 					self._actor[event.actor]))
 			return self._cursor.fetchone()[0]
 		
-		return id
+		if event.interpretation == Interpretation.MOVE_EVENT:
+			for subject in event.subjects:
+				self._cursor.execute("""
+					UPDATE event
+					SET subj_id_current=(SELECT id FROM uri WHERE value=?)
+					WHERE subj_id_current=(SELECT id FROM uri WHERE value=?)
+						AND interpretation!=? AND timestamp<?
+						""", (subject.current_uri, subject.uri,
+							self._interpretation[Interpretation.MOVE_EVENT],
+							event.timestamp))
+			# The cache has to be updated
+			for _id, c_event in self._event_cache:
+				if c_event.timestamp < event.timestamp:
+					for c_subject in c_event.subjects:
+						if c_subject.current_uri == subject.uri and \
+						not c_event.interpretation == Interpretation.MOVE_EVENT:
+							c_subject.current_uri = subject.current_uri
+		
+		return event_id
 	
 	def _store_payload (self, event):
 		# TODO: Rigth now payloads are not unique and every event has its
@@ -695,10 +831,25 @@ class ZeitgeistEngine:
 		# We can't do all(timestamps) here because the timestamps may be 0
 		if timestamps and timestamps[0] is not None and timestamps[1] is not None:
 			self._cursor.execute("DELETE FROM event WHERE id IN (%s)"
-				% ",".join(["?"] * len(ids)), ids)
+				% ",".join(str(int(_id)) for _id in ids))
 			self._cursor.connection.commit()
 			log.debug("Deleted %s" % map(int, ids))
-			
+			# thekorn: this is the keypart of the workaround for (LP: #598666)
+			# After an event got deleted we need to check the _fix_cache table
+			# for entries and remove these entries from the cache. Afterwards
+			# we have to purge all entries from the _fix_cache table.
+			clean_fix_cache = False
+			for cached_table, cached_id in self._cursor.execute("SELECT * FROM _fix_cache"):
+				if cached_table not in ("interpretation", "manifestation", "mimetype", "actor"):
+					raise ValueError("Unable to fix cache for '%s'" %cached_table)
+				clean_fix_cache = True
+				# remove deleted item from cache, for caches we are using
+				# the naming convention self._<table_name>
+				getattr(self, "_%s" %cached_table).remove_id(cached_id)
+			if clean_fix_cache:
+				# delete all rows from _fix_cache, as all caches are fixed by now
+				self._cursor.execute("DELETE FROM _fix_cache")
+				self._cursor.connection.commit()
 			self.extensions.apply_post_delete(ids, sender)
 			return timestamps
 		else:
