@@ -57,9 +57,9 @@ class RemoteInterface(SingletonApplication):
 	:const:`org.gnome.zeitgeist.Engine`.
 	"""
 	_dbus_properties = {
-		"version": DBUSProperty(lambda self: (0, 7, 1), out_signature="iii"),
+		"version": DBUSProperty(lambda self: (0, 8, 2), out_signature="iii"),
 		"extensions": DBUSProperty(
-			lambda self: list(self._engine.extensions.iter_names()),
+			lambda self: dbus.Array(self._engine.extensions.iter_names(), 's'),
 			out_signature="as"),
 	}
 	
@@ -77,7 +77,7 @@ class RemoteInterface(SingletonApplication):
 		for event in events:
 			if event is not None:
 				event._make_dbus_sendable()
-		return [NULL_EVENT if event is None else event for event in events]
+		return tuple(NULL_EVENT if event is None else event for event in events)
 	
 	# Reading stuff
 	
@@ -367,7 +367,7 @@ class RemoteInterface(SingletonApplication):
 		if self._mainloop:
 			self._mainloop.quit()
 		# remove the interface from all busses (in our case from the session bus)
-		self.remove_from_connection()
+		self._safe_quit()
 	
 	# Properties interface
 
