@@ -4,8 +4,11 @@
 #
 # Copyright © 2009-2010 Siegfried-Angel Gevatter Pujals <rainct@ubuntu.com>
 # Copyright © 2009 Mikkel Kamstrup Erlandsen <mikkel.kamstrup@gmail.com>
-# Copyright © 2009 Markus Korn <thekorn@gmx.net>
+# Copyright © 2009-2011 Markus Korn <thekorn@gmx.net>
 # Copyright © 2009 Seif Lotfy <seif@lotfy.com>
+# Copyright © 2011 J.P. Lacerda <jpaflacerda@gmail.com>
+# Copyright © 2011 Collabora Ltd.
+#             By Siegfried-Angel Gevatter Pujals <rainct@ubuntu.com>
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
@@ -24,15 +27,18 @@ import sqlite3
 import logging
 import time
 import os
+import shutil
 
 from _zeitgeist.engine import constants
 
 log = logging.getLogger("zeitgeist.sql")
 
 TABLE_MAP = {
+	"origin": "uri",
 	"subj_mimetype": "mimetype",
 	"subj_origin": "uri",
 	"subj_uri": "uri",
+	"subj_current_uri": "uri",
 }
 
 def explain_query(cursor, statement, arguments=()):
@@ -69,6 +75,14 @@ class UnicodeCursor(sqlite3.Cursor):
 			explain_query(super(UnicodeCursor, self), statement, parameters)
 		return super(UnicodeCursor, self).execute(statement, parameters)
 
+	def fetch(self, index=None):
+		if index is not None:
+			for row in self:
+				yield row[index]
+		else:
+			for row in self:
+				yield row
+
 def _get_schema_version (cursor, schema_name):
 	"""
 	Returns the schema version for schema_name or returns 0 in case
@@ -108,51 +122,87 @@ def _do_schema_upgrade (cursor, schema_name, old_version, new_version):
 	named '_zeitgeist.engine.upgrades.$schema_name_$(i)_$(i+1)' and executing 
 	the run(cursor) method of those modules until new_version is reached
 	"""
+	_do_schema_backup()
+	_set_schema_version(cursor, schema_name, -1)
 	for i in xrange(old_version, new_version):
-		# Fire of the right upgrade module
-		log.info("Upgrading database '%s' from version %s to %s. This may take a while" %
-		         (schema_name, i, i+1))
+		# Fire off the right upgrade module
+		log.info("Upgrading database '%s' from version %s to %s. "
+			"This may take a while" % (schema_name, i, i+1))
 		upgrader_name = "%s_%s_%s" % (schema_name, i, i+1)
 		module = __import__ ("_zeitgeist.engine.upgrades.%s" % upgrader_name)
 		eval("module.engine.upgrades.%s.run(cursor)" % upgrader_name)
 		
-		# Update the schema version
-		_set_schema_version(cursor, schema_name, i+1)
+	# Update the schema version
+	_set_schema_version(cursor, schema_name, new_version)
+
 	log.info("Upgrade succesful")
 
 def _check_core_schema_upgrade (cursor):
-	"""Return True if the schema is good and no setup needs to be run"""
+	"""
+	Checks whether the schema is good or, if it is outdated, triggers any
+	necessary upgrade scripts. This method will also attempt to restore a
+	database backup in case a previous upgrade was cancelled midway.
+	
+	It returns a boolean indicating whether the schema was good and the
+	database cursor (which will have changed if the database was restored).
+	"""
 	# See if we have the right schema version, and try an upgrade if needed
 	core_schema_version = _get_schema_version(cursor, constants.CORE_SCHEMA)
-	if core_schema_version is not None:
-		if core_schema_version >= constants.CORE_SCHEMA_VERSION:
-			return True
-		else:
-			try:
-				_do_schema_upgrade (cursor,
-				                    constants.CORE_SCHEMA,
-				                    core_schema_version,
-				                    constants.CORE_SCHEMA_VERSION)
-				# Don't return here. The upgrade process might depend on the
-				# tables, indexes, and views being set up (to avoid code dup)
-				log.info("Running post upgrade setup")
-				return False
-			except Exception, e:
-				log.fatal("Failed to upgrade database '%s' from version %s to %s: %s" %
-				          (constants.CORE_SCHEMA, core_schema_version, constants.CORE_SCHEMA_VERSION, e))
-				raise SystemExit(27)
+	if core_schema_version >= constants.CORE_SCHEMA_VERSION:
+		return True, cursor
 	else:
-		return False
+		try:
+			if core_schema_version <= -1:
+				cursor.connection.commit()
+				cursor.connection.close()
+				_do_schema_restore()
+				cursor = _connect_to_db(constants.DATABASE_FILE)
+				core_schema_version = _get_schema_version(cursor,
+					constants.CORE_SCHEMA)
+				log.exception("Database corrupted at upgrade -- "
+					"upgrading from version %s" % core_schema_version)
 
+			_do_schema_upgrade (cursor,
+				constants.CORE_SCHEMA,
+				core_schema_version,
+				constants.CORE_SCHEMA_VERSION)
+
+			# Don't return here. The upgrade process might depend on the
+			# tables, indexes, and views being set up (to avoid code dup)
+			log.info("Running post upgrade setup")
+			return False, cursor
+		except sqlite3.OperationalError:
+			# Something went wrong while applying the upgrade -- this is
+			# probably due to a non existing table (this occurs when 
+			# applying core_3_4, for example). We just need to fall through
+			# the rest of create_db to fix this...
+			log.exception("Database corrupted -- proceeding")
+			return False, cursor
+		except Exception, e:
+			log.exception(
+				"Failed to upgrade database '%s' from version %s to %s: %s" % \
+				(constants.CORE_SCHEMA, core_schema_version,
+				constants.CORE_SCHEMA_VERSION, e))
+			raise SystemExit(27)
+
+def _do_schema_backup ():
+	shutil.copyfile(constants.DATABASE_FILE, constants.DATABASE_FILE_BACKUP)
+
+def _do_schema_restore ():
+	shutil.move(constants.DATABASE_FILE_BACKUP, constants.DATABASE_FILE)
+
+def _connect_to_db(file_path):
+	conn = sqlite3.connect(file_path)
+	conn.row_factory = sqlite3.Row
+	cursor = conn.cursor(UnicodeCursor)
+	return cursor
 
 def create_db(file_path):
 	"""Create the database and return a default cursor for it"""
 	start = time.time()
 	log.info("Using database: %s" % file_path)
 	new_database = not os.path.exists(file_path)
-	conn = sqlite3.connect(file_path)
-	conn.row_factory = sqlite3.Row
-	cursor = conn.cursor(UnicodeCursor)
+	cursor = _connect_to_db(file_path)
 
 	# Seif: as result of the optimization story (LP: #639737) we are setting
 	# journal_mode to WAL if possible, this change is irreversible but
@@ -165,9 +215,19 @@ def create_db(file_path):
 	# one connection to the database is allowed to revert this setting set locking_mode to NORMAL.
 	cursor.execute("PRAGMA locking_mode = EXCLUSIVE")
 	
+	# thekorn: as part of the workaround for (LP: #598666) we need to
+	# create the '_fix_cache' TEMP table on every start,
+	# this table gets purged once the engine gets closed.
+	# When a cached value gets deleted we automatically store the name
+	# of the cache and the value's id to this table. It's then up to
+	# the python code to delete items from the cache based on the content
+	# of this table.
+	cursor.execute("CREATE TEMP TABLE _fix_cache (table_name VARCHAR, id INTEGER)")
+	
 	# Always assume that temporary memory backed DBs have good schemas
-	if constants.DATABASE_FILE != ":memory:":
-		if not new_database and _check_core_schema_upgrade(cursor):
+	if constants.DATABASE_FILE != ":memory:" and not new_database:
+		do_upgrade, cursor = _check_core_schema_upgrade(cursor)
+		if do_upgrade:
 			_time = (time.time() - start)*1000
 			log.debug("Core schema is good. DB loaded in %sms" % _time)
 			return cursor
@@ -242,24 +302,25 @@ def create_db(file_path):
 		CREATE TABLE IF NOT EXISTS storage
 			(id INTEGER PRIMARY KEY,
 			 value VARCHAR UNIQUE,
-			 state INTEGER)
+			 state INTEGER,
+			 icon VARCHAR,
+			 display_name VARCHAR)
 		""")
 	cursor.execute("""
 		CREATE UNIQUE INDEX IF NOT EXISTS storage_value
 			ON storage(value)""")
 	
 	# event - the primary table for log statements
-	# note that event.id is NOT unique, we can have multiple subjects per id
-	# timestamps are integers (for now), if you would like to change it
-	# please start a bugreport for it. In case we agree on this change
-	# remember to also fix our unittests to reflect this change
+	#  - Note that event.id is NOT unique, we can have multiple subjects per ID
+	#  - Timestamps are integers.
+	#  - (event-)origin and subj_id_current are added to the end of the table
 	cursor.execute("""
 		CREATE TABLE IF NOT EXISTS event (
 			id INTEGER,
 			timestamp INTEGER,
 			interpretation INTEGER,
-			manifestation INTEGER,			 
-			actor INTEGER,			 
+			manifestation INTEGER,
+			actor INTEGER,
 			payload INTEGER,
 			subj_id INTEGER,
 			subj_interpretation INTEGER,
@@ -268,16 +329,22 @@ def create_db(file_path):
 			subj_mimetype INTEGER,
 			subj_text INTEGER,
 			subj_storage INTEGER,
+			origin INTEGER,
+			subj_id_current INTEGER,
 			CONSTRAINT interpretation_fk FOREIGN KEY(interpretation)
 				REFERENCES interpretation(id) ON DELETE CASCADE,
 			CONSTRAINT manifestation_fk FOREIGN KEY(manifestation)
 				REFERENCES manifestation(id) ON DELETE CASCADE,
 			CONSTRAINT actor_fk FOREIGN KEY(actor)
 				REFERENCES actor(id) ON DELETE CASCADE,
+			CONSTRAINT origin_fk FOREIGN KEY(origin)
+				REFERENCES uri(id) ON DELETE CASCADE,
 			CONSTRAINT payload_fk FOREIGN KEY(payload)
 				REFERENCES payload(id) ON DELETE CASCADE,
 			CONSTRAINT subj_id_fk FOREIGN KEY(subj_id)
 				REFERENCES uri(id) ON DELETE CASCADE,
+			CONSTRAINT subj_id_current_fk FOREIGN KEY(subj_id_current)
+				REFERENCES uri(id) ON DELETE CASCADE,
 			CONSTRAINT subj_interpretation_fk FOREIGN KEY(subj_interpretation)
 				REFERENCES interpretation(id) ON DELETE CASCADE,
 			CONSTRAINT subj_manifestation_fk FOREIGN KEY(subj_manifestation)
@@ -308,9 +375,15 @@ def create_db(file_path):
 	cursor.execute("""
 		CREATE INDEX IF NOT EXISTS event_actor
 			ON event(actor)""")
+	cursor.execute("""
+		CREATE INDEX IF NOT EXISTS event_origin
+			ON event(origin)""")
 	cursor.execute("""
 		CREATE INDEX IF NOT EXISTS event_subj_id
 			ON event(subj_id)""")
+	cursor.execute("""
+		CREATE INDEX IF NOT EXISTS event_subj_id_current
+			ON event(subj_id_current)""")
 	cursor.execute("""
 		CREATE INDEX IF NOT EXISTS event_subj_interpretation
 			ON event(subj_interpretation)""")
@@ -351,24 +424,25 @@ def create_db(file_path):
 				""" % {'column': column, 'table': table})
 
 	# ... special cases
-	cursor.execute("""
-		CREATE TRIGGER IF NOT EXISTS fkdc_event_uri_1
-		BEFORE DELETE ON event
-		WHEN ((SELECT COUNT(*) FROM event WHERE subj_id=OLD.subj_id OR subj_origin=OLD.subj_id) < 2)
-		BEGIN
-			DELETE FROM uri WHERE id=OLD.subj_id;
-		END;
-		""" % {'column': column, 'table': table})
-	cursor.execute("""
-		CREATE TRIGGER IF NOT EXISTS fkdc_event_uri_2
-		BEFORE DELETE ON event
-		WHEN ((SELECT COUNT(*) FROM event WHERE subj_id=OLD.subj_origin OR subj_origin=OLD.subj_origin) < 2)
-		BEGIN
-			DELETE FROM uri WHERE id=OLD.subj_origin;
-		END;
-		""" % {'column': column, 'table': table})
+	for num, column in enumerate(('subj_id', 'subj_origin',
+	'subj_id_current', 'origin')):
+		cursor.execute("""
+			CREATE TRIGGER IF NOT EXISTS fkdc_event_uri_%(num)d
+			BEFORE DELETE ON event
+			WHEN ((
+				SELECT COUNT(*)
+				FROM event
+				WHERE
+					origin=OLD.%(column)s
+					OR subj_id=OLD.%(column)s
+					OR subj_id_current=OLD.%(column)s
+					OR subj_origin=OLD.%(column)s
+				) < 2)
+			BEGIN
+				DELETE FROM uri WHERE id=OLD.%(column)s;
+			END;
+			""" % {'num': num+1, 'column': column})
 
-	# TODO: Make the DROP conditional to version upgrades. How?
 	cursor.execute("DROP VIEW IF EXISTS event_view")
 	cursor.execute("""
 		CREATE VIEW IF NOT EXISTS event_view AS
@@ -393,7 +467,13 @@ def create_db(file_path):
 				(SELECT value FROM storage
 					WHERE storage.id=event.subj_storage) AS subj_storage,
 				(SELECT state FROM storage
-					WHERE storage.id=event.subj_storage) AS subj_storage_state
+					WHERE storage.id=event.subj_storage) AS subj_storage_state,
+				event.origin,
+				(SELECT value FROM uri WHERE uri.id=event.origin)
+					AS event_origin_uri,
+				(SELECT value FROM uri WHERE uri.id=event.subj_id_current)
+					AS subj_current_uri,
+				event.subj_id_current
 			FROM event
 		""")
 	
@@ -432,6 +512,14 @@ class TableLookup(dict):
 			self[row["value"]] = row["id"]
 		
 		self._inv_dict = dict((value, key) for key, value in self.iteritems())
+		
+		cursor.execute("""
+			CREATE TEMP TRIGGER update_cache_%(table)s
+			BEFORE DELETE ON %(table)s
+			BEGIN
+				INSERT INTO _fix_cache VALUES ("%(table)s", OLD.id);
+			END;
+			""" % {"table": table})
 	
 	def __getitem__(self, name):
 		# Use this for inserting new properties into the database
@@ -462,6 +550,11 @@ class TableLookup(dict):
 		# database already. Eg., in find_eventids.
 		return super(TableLookup, self).__getitem__(name)
 		
+	def remove_id(self, id):
+		value = self.value(id)
+		del self._inv_dict[id]
+		del self[value]
+		
 def get_right_boundary(text):
 	""" returns the smallest string which is greater than `text` """
 	if not text:
@@ -526,28 +619,36 @@ class WhereClause:
 	def __len__(self):
 		return len(self._conditions)
 	
-	def add(self, condition, arguments):
+	def add(self, condition, arguments=None):
 		if not condition:
 			return
 		self._conditions.append(condition)
-		if not hasattr(arguments, "__iter__"):
-			self.arguments.append(arguments)
-		else:
-			self.arguments.extend(arguments)
+		if arguments is not None:
+			if not hasattr(arguments, "__iter__"):
+				self.arguments.append(arguments)
+			else:
+				self.arguments.extend(arguments)
 			
 	def add_text_condition(self, column, value, like=False, negation=False, cache=None):
 		if like:
-			assert column in ("subj_uri", "subj_origin", "actor", "subj_mimetype"), \
-				"prefix search on the %r column is not supported by zeitgeist" %column
+			assert column in ("origin", "subj_uri", "subj_current_uri",
+			"subj_origin", "actor", "subj_mimetype"), \
+				"prefix search on the %r column is not supported by zeitgeist" % column
 			if column == "subj_uri":
 				# subj_id directly points to the id of an uri entry
 				view_column = "subj_id"
+			elif column == "subj_current_uri":
+				view_column = "subj_id_current"
 			else:
 				view_column = column
 			optimized_glob, value = self.optimize_glob("id", TABLE_MAP.get(column, column), value)
 			sql = "%s %sIN (%s)" %(view_column, self.NOT if negation else "", optimized_glob)
+			if negation:
+				sql += " OR %s IS NULL" % view_column
 		else:
-			if column == "subj_origin":
+			if column == "origin":
+				column ="event_origin_uri"
+			elif column == "subj_origin":
 				column = "subj_origin_uri"
 			sql = "%s %s= ?" %(column, "!" if negation else "")
 			if cache is not None:
