@@ -84,7 +84,7 @@ class SingletonApplication(dbus.service.Object):
 			else:
 				logging.debug("Replacing currently running process...")
 			try:
-				interface = ZeitgeistDBusInterface()
+				interface = ZeitgeistDBusInterface(reconnect=False)
 				interface.Quit()
 				while dbus.SessionBus().name_has_owner(ZeitgeistDBusInterface.BUS_NAME):
 					pass
@@ -97,3 +97,13 @@ class SingletonApplication(dbus.service.Object):
 		else:
 			raise RuntimeError("An existing instance was found. Please use " \
 				 "--replace to quit it and start a new instance.")
+				 
+	def _safe_quit(self):
+		# safely quit the interface on the bus by removing this interface
+		# from the bus, and releasing the (by-hand) registered bus name
+		try:
+			self.remove_from_connection()
+			self.connection.release_name(ZeitgeistDBusInterface.BUS_NAME)
+		except Exception, e:
+			log.error("Could not remove singleton properly due to the following error: %s"
+					%e)
