@@ -47,6 +47,7 @@ class _DBusInterface(object):
 	# that here because otherwise all instances would share their state.
 	_disconnect_callbacks = None
 	_reconnect_callbacks = None
+	_generic_callbacks = None
 
 	@staticmethod
 	def get_members(introspection_xml):
@@ -66,14 +67,18 @@ class _DBusInterface(object):
 		return methods, signals
 
 	def reconnect(self):
-		self.__proxy = dbus.SessionBus().get_object(self.__iface.requested_bus_name,
-			self.__object_path)
+		if not self._reconnect_when_needed:
+			return
+		self.__proxy = dbus.SessionBus().get_object(
+			self.__iface.requested_bus_name, self.__object_path)
 		self.__iface = dbus.Interface(self.__proxy, self.__interface_name)
+		self._load_introspection_data()
 
-	def _disconnection_safe(self, meth, *args, **kwargs):
+	def _disconnection_safe(self, method_getter, *args, **kwargs):
 		"""
-		Executes the given method. If it fails because the D-Bus connection
-		was lost, attempts to recover it and executes the method again.
+		Executes the method returned by method_getter. If it fails because
+		the D-Bus connection was lost, it attempts to recover it and executes
+		the method again.
 		"""
 		
 		custom_error_handler = None
@@ -85,7 +90,7 @@ class _DBusInterface(object):
 				self.reconnect()
 				# We don't use the reconnecting_error_handler here since that'd
 				# get us into an endless loop if Zeitgeist really isn't there.
-				return meth(*args, **original_kwargs)
+				return method_getter()(*args, **original_kwargs)
 			else:
 				if custom_error_handler is not None:
 					custom_error_handler(e)
@@ -99,26 +104,34 @@ class _DBusInterface(object):
 			kwargs['error_handler'] = reconnecting_error_handler
 
 		try:
-			return meth(*args, **kwargs)
+			return method_getter()(*args, **kwargs)
 		except dbus.exceptions.DBusException, e:
-			reconnecting_error_handler()
+			return reconnecting_error_handler(e)
 
 	def __getattr__(self, name):
-		if name not in self.__methods:
+		if self.__methods is not None and name not in self.__methods:
 			raise TypeError("Unknown method name: %s" % name)
 		def _ProxyMethod(*args, **kwargs):
 			"""
 			Method wrapping around a D-Bus call, which attempts to recover
 			the connection to Zeitgeist if it got lost.
 			"""
-			return self._disconnection_safe(getattr(self.__iface, name),
-				*args, **kwargs)
+			return self._disconnection_safe(
+				lambda: getattr(self.__iface, name), *args, **kwargs)
 		return _ProxyMethod
 
+	def get_property(self, property_name):
+		return self._disconnection_safe(
+			lambda: self.__proxy.get_dbus_method("Get", dbus.PROPERTIES_IFACE),
+			self.__interface_name, property_name)
+
 	def connect(self, signal, callback, **kwargs):
 		"""Connect a callback to a signal of the current proxy instance."""
+		if self.__signals is None:
+			self.reconnect()
 		if signal not in self.__signals:
 			raise TypeError("Unknown signal name: %s" % signal)
+		self._generic_callbacks.add((signal, callback))
 		self.__proxy.connect_to_signal(
 			signal,
 			callback,
@@ -141,23 +154,39 @@ class _DBusInterface(object):
 	def proxy(self):
 		return self.__proxy
 
-	def __init__(self, proxy, interface_name, object_path):
+	def _load_introspection_data(self):
+		self.__methods, self.__signals = self.get_members(
+			self.__proxy.Introspect(
+				dbus_interface='org.freedesktop.DBus.Introspectable'))
+
+	def __init__(self, proxy, interface_name, object_path, reconnect=True):
 		self.__proxy = proxy
 		self.__interface_name = interface_name
 		self.__object_path = object_path
 		self.__iface = dbus.Interface(proxy, interface_name)
-		self.__methods, self.__signals = self.get_members(proxy.Introspect())
+		self._reconnect_when_needed = reconnect
+		self._load_introspection_data()
 		
 		self._disconnect_callbacks = set()
 		self._reconnect_callbacks = set()
+		self._generic_callbacks = set()
 		
 		# Listen to (dis)connection notifications, for connect_exit and connect_join
 		def name_owner_changed(connection_name):
 			if connection_name == "":
 				callbacks = self._disconnect_callbacks
+				self.__methods = self.__signals = None
 			else:
+				if not self._reconnect_when_needed:
+					return
 				self.reconnect()
 				callbacks = self._reconnect_callbacks
+				for signal, callback in self._generic_callbacks:
+					try:
+						self.connect(signal, callback)
+					except TypeError:
+						log.exception("Failed to reconnect to signal \"%s\" "
+							"after engine disconnection." % signal)
 			for callback in callbacks:
 				callback()
 		dbus.SessionBus().watch_name_owner(self.__iface.requested_bus_name,
@@ -183,18 +212,12 @@ class ZeitgeistDBusInterface(object):
 	def version(self):
 		"""Returns the API version"""
 		dbus_interface = self.__shared_state["dbus_interface"]
-		return dbus_interface._disconnection_safe(
-			dbus_interface.proxy.get_dbus_method("Get",
-				dbus_interface=dbus.PROPERTIES_IFACE),
-			self.INTERFACE_NAME, "version")
+		return dbus_interface.get_property("version")
 	
 	def extensions(self):
 		"""Returns active extensions"""
 		dbus_interface = self.__shared_state["dbus_interface"]
-		return dbus_interface._disconnection_safe(
-			dbus_interface.proxy.get_dbus_method("Get",
-				dbus_interface=dbus.PROPERTIES_IFACE),
-				self.INTERFACE_NAME, "extensions")
+		return dbus_interface.get_property("extensions")
 	
 	def get_extension(cls, name, path, busname=None):
 		""" Returns an interface to the given extension.
@@ -218,7 +241,7 @@ class ZeitgeistDBusInterface(object):
 			cls.__shared_state["extension_interfaces"][name] = iface
 		return cls.__shared_state["extension_interfaces"][name]
 	
-	def __init__(self):
+	def __init__(self, reconnect=True):
 		if not "dbus_interface" in self.__shared_state:
 			try:
 				proxy = dbus.SessionBus().get_object(self.BUS_NAME,
@@ -232,7 +255,7 @@ class ZeitgeistDBusInterface(object):
 					raise
 			self.__shared_state["extension_interfaces"] = {}
 			self.__shared_state["dbus_interface"] = _DBusInterface(proxy,
-				self.INTERFACE_NAME, self.OBJECT_PATH)
+				self.INTERFACE_NAME, self.OBJECT_PATH, reconnect)
 
 class Monitor(dbus.service.Object):
 	"""
@@ -249,21 +272,27 @@ class Monitor(dbus.service.Object):
 	
 	# Used in Monitor._next_path() to generate unique path names
 	_last_path_id = 0
+	
+	_event_type = Event
 
 	def __init__ (self, time_range, event_templates, insert_callback,
-		delete_callback, monitor_path=None):
+		delete_callback, monitor_path=None, event_type=None):
 		if not monitor_path:
 			monitor_path = Monitor._next_path()
 		elif isinstance(monitor_path, (str, unicode)):
 			monitor_path = dbus.ObjectPath(monitor_path)
 		
+		if event_type:
+			if not issubclass(event_type, Event):
+				raise TypeError("Event subclass expected.")
+			self._event_type = event_type
+		
 		self._time_range = time_range
 		self._templates = event_templates
 		self._path = monitor_path
 		self._insert_callback = insert_callback
 		self._delete_callback = delete_callback
 		dbus.service.Object.__init__(self, dbus.SessionBus(), monitor_path)
-
 	
 	def get_path (self): return self._path
 	path = property(get_path,
@@ -295,7 +324,7 @@ class Monitor(dbus.service.Object):
 		    See :meth:`ZeitgeistClient.install_monitor`
 		"""
 		self._insert_callback(TimeRange(time_range[0], time_range[1]),
-			map(Event, events))
+			map(self._event_type, events))
 	
 	@dbus.service.method("org.gnome.zeitgeist.Monitor",
 	                     in_signature="(xx)au")
@@ -342,6 +371,7 @@ class ZeitgeistClient:
 	"""
 	
 	_installed_monitors = []
+	_event_type = Event
 	
 	@staticmethod
 	def get_event_and_extra_arguments(arguments):
@@ -374,6 +404,35 @@ class ZeitgeistClient:
 						"Error reinstalling monitor: %s" % err))
 		self._iface.connect_join(reconnect_monitors)
 	
+	def register_event_subclass(self, event_type):
+		"""
+		Register a subclass of Event with this ZeiteistClient instance. When
+		data received over D-Bus is instantiated into an Event object, the
+		provided subclass will be used.
+		"""
+		if not issubclass(event_type, Event):
+			raise TypeError("Event subclass expected.")
+		self._event_type = event_type
+	
+	def register_subject_subclass(self, subject_type):
+		"""
+		Register a subclass of Subject with this ZeiteistClient instance. When
+		data received over D-Bus is instantiated into a Subject object, the
+		provided subclass will be used.
+		
+		Note that this method works by changing the Event type associated with
+		this ZeitgeistClient instance, so it should always be called *after*
+		any register_event_subclass calls.
+		
+		Even better, if you also have a custom Event subclass, you may directly
+		override the Subject type by changing its _subject_type class variable.
+		"""
+		if not issubclass(subject_type, Subject):
+			raise TypeError("Subject subclass expected.")
+		class EventWithCustomSubject(self._event_type):
+			_subject_type = subject_type
+		self._event_type = EventWithCustomSubject
+	
 	def _safe_error_handler(self, error_handler, *args):
 		if error_handler is not None:
 			if callable(error_handler):
@@ -656,7 +715,7 @@ class ZeitgeistClient:
 					num_events,
 					result_type,
 					reply_handler=lambda raw: events_reply_handler(
-						map(Event.new_for_struct, raw)),
+						map(self._event_type.new_for_struct, raw)),
 					error_handler=self._safe_error_handler(error_handler,
 						events_reply_handler, []))
 	
@@ -717,7 +776,7 @@ class ZeitgeistClient:
 		# the raw DBus reply into a list of Event instances
 		self._iface.GetEvents(event_ids,
 				reply_handler=lambda raw: events_reply_handler(
-					map(Event.new_for_struct, raw)),
+					map(self._event_type.new_for_struct, raw)),
 				error_handler=self._safe_error_handler(error_handler,
 						events_reply_handler, []))
 	
@@ -868,7 +927,8 @@ class ZeitgeistClient:
 		
 		
 		mon = Monitor(time_range, event_templates, notify_insert_handler,
-			notify_delete_handler, monitor_path=monitor_path)
+			notify_delete_handler, monitor_path=monitor_path,
+			event_type=self._event_type)
 		self._iface.InstallMonitor(mon.path,
 		                           mon.time_range,
 		                           mon.templates,
