@@ -6,6 +6,8 @@
 # Copyright © 2009 Markus Korn <thekorn@gmx.de>
 # Copyright © 2009-2010 Seif Lotfy <seif@lotfy.com>
 # Copyright © 2009-2010 Siegfried-Angel Gevatter Pujals <rainct@ubuntu.com>
+# Copyright © 2011 Collabora Ltd.
+#             By Siegfried-Angel Gevatter Pujals <rainct@ubuntu.com>
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
@@ -85,62 +87,18 @@ def get_timestamp_for_now():
 	Return the current time in milliseconds since the Unix Epoch.
 	"""
 	return int(time.time() * 1000)
-		
-
-class enum_factory(object):
-	"""factory for enums"""
-	counter = 0
-	
-	def __init__(self, doc):
-		self.__doc__ = doc
-		self._id = enum_factory.counter
-		enum_factory.counter += 1
-		
-		
-class EnumMeta(type):
-	"""Metaclass to register enums in correct order and assign interger
-	values to them
-	"""
-	def __new__(cls, name, bases, attributes):
-		enums = filter(
-			lambda x: isinstance(x[1], enum_factory), attributes.iteritems()
-		)
-		enums = sorted(enums, key=lambda x: x[1]._id)
-		for n, (key, value) in enumerate(enums):
-			attributes[key] = EnumValue(n, value.__doc__)
-		return super(EnumMeta, cls).__new__(cls, name, bases, attributes)
-
 
 class EnumValue(int):
-	"""class which behaves like an int, but has an additional docstring"""
+	"""Class which behaves like an int, but has an additional docstring"""
+	
 	def __new__(cls, value, doc=""):
 		obj = super(EnumValue, cls).__new__(EnumValue, value)
 		obj.__doc__ = "%s. ``(Integer value: %i)``" %(doc, obj)
 		return obj
 
-
-class Enum(object):
-
-	def __init__(self, docstring):
-		self.__doc__ = str(docstring)
-		self.__enums = {}
-
-	def __getattr__(self, name):
-		try:
-			return self.__enums[name]
-		except KeyError:
-			raise AttributeError
-
-	def register(self, value, name, docstring):
-		ids = map(int, self.__enums.values())
-		if value in ids or name in self.__enums:
-			raise ValueError
-		self.__enums[name] = EnumValue(value, docstring)
-		
-		
 def isCamelCase(text):
 	return text and text[0].isupper() and " " not in text
-	
+
 def get_name_or_str(obj):
 	try:
 		return str(obj.name)
@@ -274,7 +232,6 @@ class Symbol(str):
 				parent = _SYMBOLS_BY_URI[parent]
 			except KeyError, e:
 				# Parent is not a known URI
-				#print 11111111111, self.uri, parent #debug output
 				return self.uri == parent
 		
 		# Invariant: parent is a Symbol
@@ -371,6 +328,16 @@ class TimeRange(list):
 		now = int(time.time() * 1000)
 		return cls(now - (sec * 1000), now)
 	
+	@classmethod
+	def from_timestamp(cls, timestamp):
+		"""
+		Return a :class:`TimeRange` ranging from the given timestamp until
+		the end of time.
+		
+		The given timestamp is expected to be expressed in miliseconds.
+		"""
+		return cls(int(timestamp), cls._max_stamp)
+	
 	@classmethod
 	def always(cls):
 		"""
@@ -417,19 +384,8 @@ class TimeRange(list):
 				result.end = time_range.end
 		
 		return result
-		
-		
-class RelevantResultType(object):
-	"""
-	An enumeration class used to define how query results should be returned
-	from the Zeitgeist engine.
-	"""
-	__metaclass__ = EnumMeta
-	
-	Recent = enum_factory("All uris with the most recent uri first")
-	Related = enum_factory("All uris with the most related one first")
-	
-	
+
+
 class Subject(list):
 	"""
 	Represents a subject of an :class:`Event`. This class is both used to
@@ -445,18 +401,21 @@ class Subject(list):
 		Origin,
 		Mimetype,
 		Text,
-		Storage) = range(7)
+		Storage,
+		CurrentUri) = range(8)
 		
-	SUPPORTS_NEGATION = (Uri, Interpretation, Manifestation, Origin, Mimetype)
-	SUPPORTS_WILDCARDS = (Uri, Origin, Mimetype)
+	SUPPORTS_NEGATION = (Uri, CurrentUri, Interpretation, Manifestation, Origin, Mimetype)
+	SUPPORTS_WILDCARDS = (Uri, CurrentUri, Origin, Mimetype)
 	
 	def __init__(self, data=None):
-		super(Subject, self).__init__([""]*len(Subject.Fields))
 		if data:
+			if len(data) == len(Subject.Fields) - 1:
+				# current_uri has been added in Zeitgeist 0.8.0
+				data.append("")
 			if len(data) != len(Subject.Fields):
 				raise ValueError(
-					"Invalid subject data length %s, expected %s"
-					% (len(data), len(Subject.Fields)))
+					"Invalid subject data length %s, expected %s" \
+					%(len(data), len(Subject.Fields)))
 			super(Subject, self).__init__(data)
 		else:
 			super(Subject, self).__init__([""]*len(Subject.Fields))
@@ -466,6 +425,15 @@ class Subject(list):
 			self.__class__.__name__, super(Subject, self).__repr__()
 		)
 	
+	def __eq__(self, other):
+		for field in Subject.Fields:
+			if field is Subject.CurrentUri and not self[field] or \
+			not other[field]:
+				continue
+			if self[field] != other[field]:
+				return False
+		return True
+	
 	@staticmethod
 	def new_for_values (**values):
 		"""
@@ -473,6 +441,7 @@ class Subject(list):
 		to the keyword arguments passed to this method.
 		
 		:param uri: The URI of the subject. Eg. *file:///tmp/ratpie.txt*
+		:param current_uri: The current known URI of the subject (if it was moved or deleted).
 		:param interpretation: The interpretation type of the subject, given either as a string URI or as a :class:`Interpretation` instance
 		:param manifestation: The manifestation type of the subject, given either as a string URI or as a :class:`Manifestation` instance
 		:param origin: The URI of the location where subject resides or can be found
@@ -482,7 +451,7 @@ class Subject(list):
 		"""
 		self = Subject()
 		for key, value in values.iteritems():
-			if not key in ("uri", "interpretation", "manifestation", "origin",
+			if not key in ("uri", "current_uri", "interpretation", "manifestation", "origin",
 						"mimetype", "text", "storage"):
 				raise ValueError("Subject parameter '%s' is not supported" %key)
 			setattr(self, key, value)
@@ -495,6 +464,14 @@ class Subject(list):
 		self[Subject.Uri] = value
 	uri = property(get_uri, set_uri,
 	doc="Read/write property with the URI of the subject encoded as a string")
+	
+	def get_current_uri(self):
+		return self[Subject.CurrentUri]
+	
+	def set_current_uri(self, value):
+		self[Subject.CurrentUri] = value
+	current_uri = property(get_current_uri, set_current_uri,
+	doc="Read/write property with the current URI of the subject encoded as a string")
 		
 	def get_interpretation(self):
 		return self[Subject.Interpretation]
@@ -602,10 +579,13 @@ class Event(list):
 		Timestamp,
 		Interpretation,
 		Manifestation,
-		Actor) = range(5)
-		
-	SUPPORTS_NEGATION = (Interpretation, Manifestation, Actor)
-	SUPPORTS_WILDCARDS = (Actor,)
+		Actor,
+		Origin) = range(6)
+	
+	SUPPORTS_NEGATION = (Interpretation, Manifestation, Actor, Origin)
+	SUPPORTS_WILDCARDS = (Actor, Origin)
+	
+	_subject_type = Subject
 	
 	def __init__(self, struct = None):
 		"""
@@ -633,26 +613,44 @@ class Event(list):
 		super(Event, self).__init__()
 		if struct:
 			if len(struct) == 1:
-				self.append(struct[0])
+				self.append(self._check_event_struct(struct[0]))
 				self.append([])
 				self.append("")
 			elif len(struct) == 2:
-				self.append(struct[0])
-				self.append(map(Subject, struct[1]))
+				self.append(self._check_event_struct(struct[0]))
+				self.append(map(self._subject_type, struct[1]))
 				self.append("")
 			elif len(struct) == 3:
-				self.append(struct[0])
-				self.append(map(Subject, struct[1]))
+				self.append(self._check_event_struct(struct[0]))
+				self.append(map(self._subject_type, struct[1]))
 				self.append(struct[2])
 			else:
 				raise ValueError("Invalid struct length %s" % len(struct))
+			# If this event is being created from an existing Event instance,
+			# make a copy of the list holding the event information. This
+			# enables the idiom "event2 = Event(event1)" to copy an event.
+			if isinstance(struct, Event):
+				self[0] = list(self[0])
 		else:
 			self.extend(([""]* len(Event.Fields), [], ""))
 		
 		# If we have no timestamp just set it to now
 		if not self[0][Event.Timestamp]:
 			self[0][Event.Timestamp] = str(get_timestamp_for_now())
-		
+		# If we have no origin for Event then we set None
+		if len(self[0]) == 5:
+			self[0].append(None)
+	
+	@classmethod
+	def _check_event_struct(cls, event_data):
+		if len(event_data) == len(cls.Fields) - 1:
+			# Old versions of Zeitgeist didn't have the event origin field.
+			event_data.append("")
+		if len(event_data) != len(cls.Fields):
+			raise ValueError("event_data must have %s members, found %s" % \
+				(len(cls.Fields), len(event_data)))
+		return event_data
+	
 	@classmethod
 	def new_for_data(cls, event_data):
 		"""
@@ -662,10 +660,7 @@ class Event(list):
 		Event.Fields enumeration.
 		"""
 		self = cls()
-		if len(event_data) != len(cls.Fields):
-			raise ValueError("event_data must have %s members, found %s" % \
-				(len(cls.Fields), len(event_data)))
-		self[0] = event_data
+		self[0] = self._check_event_struct(event_data)
 		return self
 		
 	@classmethod
@@ -685,7 +680,8 @@ class Event(list):
 		:param timestamp: Event timestamp in milliseconds since the Unix Epoch 
 		:param interpretaion: The Interpretation type of the event
 		:param manifestation: Manifestation type of the event
-		:param actor: The actor (application) that triggered the event
+		:param actor: The actor (application) that triggered the event		
+		:param origin: The origin (domain) where the event was triggered
 		:param subjects: A list of :class:`Subject` instances
 		
 		Instead of setting the *subjects* argument one may use a more
@@ -694,19 +690,19 @@ class Event(list):
 		counterparts in :meth:`Subject.new_for_values`:
 		
 		:param subject_uri:
+		:param subject_current_uri:
 		:param subject_interpretation:
 		:param subject_manifestation:
 		:param subject_origin:
 		:param subject_mimetype:
 		:param subject_text:
 		:param subject_storage:
-		 
-		
 		"""
 		self = cls()
 		for key in values:
 			if not key in ("timestamp", "interpretation", "manifestation",
-				"actor", "subjects", "subject_uri", "subject_interpretation",
+				"actor", "origin", "subjects", "subject_uri",
+				"subject_current_uri", "subject_interpretation",
 				"subject_manifestation", "subject_origin", "subject_mimetype",
 				"subject_text", "subject_storage"):
 				raise ValueError("Event parameter '%s' is not supported" % key)
@@ -715,13 +711,15 @@ class Event(list):
 		self.interpretation = values.get("interpretation", "")
 		self.manifestation = values.get("manifestation", "")
 		self.actor = values.get("actor", "")
+		self.origin = values.get("origin", "")
 		self.subjects = values.get("subjects", self.subjects)
 		
 		if self._dict_contains_subject_keys(values):
 			if "subjects" in values:
 				raise ValueError("Subject keys, subject_*, specified together with full subject list")
-			subj = Subject()
+			subj = self._subject_type()
 			subj.uri = values.get("subject_uri", "")
+			subj.current_uri = values.get("subject_current_uri", "")
 			subj.interpretation = values.get("subject_interpretation", "")
 			subj.manifestation = values.get("subject_manifestation", "")
 			subj.origin = values.get("subject_origin", "")
@@ -734,13 +732,14 @@ class Event(list):
 	
 	@staticmethod
 	def _dict_contains_subject_keys (dikt):
-		if "subject_uri" in dikt : return True
-		elif "subject_interpretation" in dikt : return True
-		elif "subject_manifestation" in dikt : return True
-		elif "subject_origin" in dikt : return True
-		elif "subject_mimetype" in dikt : return True
-		elif "subject_text" in dikt : return True
-		elif "subject_storage" in dikt : return True
+		if "subject_uri" in dikt: return True
+		elif "subject_current_uri" in dikt: return True
+		elif "subject_interpretation" in dikt: return True
+		elif "subject_manifestation" in dikt: return True
+		elif "subject_origin" in dikt: return True
+		elif "subject_mimetype" in dikt: return True
+		elif "subject_text" in dikt: return True
+		elif "subject_storage" in dikt: return True
 		return False
 	
 	def __repr__(self):
@@ -753,12 +752,12 @@ class Event(list):
 		Append a new empty Subject and return a reference to it
 		"""
 		if not subject:
-			subject = Subject()
+			subject = self._subject_type()
 		self.subjects.append(subject)
 		return subject
 	
 	def get_subjects(self):
-		return self[1]	
+		return self[1]
 	
 	def set_subjects(self, subjects):
 		self[1] = subjects
@@ -801,7 +800,20 @@ class Event(list):
 	def set_actor(self, value):
 		self[0][Event.Actor] = value
 	actor = property(get_actor, set_actor,
-	doc="Read/write property defining the application or entity responsible for emitting the event. For applications the format of this field is base filename of the corresponding .desktop file with an `app://` URI scheme. For example `/usr/share/applications/firefox.desktop` is encoded as `app://firefox.desktop`")
+	doc="Read/write property defining the application or entity responsible "
+		"for emitting the event. For applications, the format of this field is "
+		"the base filename of the corresponding .desktop file with an "
+		"`application://` URI scheme. For example, "
+		"`/usr/share/applications/firefox.desktop` is encoded as "
+		"`application://firefox.desktop`")
+	
+	def get_origin(self):
+		return self[0][Event.Origin]
+	
+	def set_origin(self, value):
+		self[0][Event.Origin] = value
+	origin = property(get_origin, set_origin,
+	doc="Read/write property defining the origin where the event was emitted.")
 	
 	def get_payload(self):
 		return self[2]
@@ -979,17 +991,35 @@ class DataSource(list):
 	def __repr__(self):
 		return "%s: %s (%s)" % (self.__class__.__name__, self[self.UniqueId],
 			self[self.Name])
-	
-	
+
 
 NULL_EVENT = ([], [], [])
 """Minimal Event representation, a tuple containing three empty lists.
 This `NULL_EVENT` is used by the API to indicate a queried but not
 available (not found or blocked) Event.
 """
-		
-		
-class StorageState(object):
+
+class _Enumeration(object):
+
+	@classmethod
+	def iteritems(self):
+		"""
+		Return an iterator yielding (name, value) tuples for all items in
+		this enumeration.
+		"""
+		return iter(map(lambda x: (x, getattr(self, x)),
+			filter(lambda x: not x.startswith('__'), sorted(self.__dict__))))
+
+class RelevantResultType(_Enumeration):
+	"""
+	An enumeration class used to define how query results should be returned
+	from the Zeitgeist engine.
+	"""
+	
+	Recent = EnumValue(0, "All uris with the most recent uri first")
+	Related = EnumValue(1, "All uris with the most related one first")
+
+class StorageState(_Enumeration):
 	"""
 	Enumeration class defining the possible values for the storage state
 	of an event subject.
@@ -999,62 +1029,94 @@ class StorageState(object):
 	deleted files, files on unplugged USB drives, files available only when
 	a network is available etc.
 	"""
-	__metaclass__ = EnumMeta
 	
-	NotAvailable = enum_factory(("The storage medium of the events "
-		"subjects must not be available to the user"))
-	Available = enum_factory(("The storage medium of all event subjects "
-		"must be immediately available to the user"))
-	Any = enum_factory("The event subjects may or may not be available")
+	NotAvailable = EnumValue(0, "The storage medium of the events "
+		"subjects must not be available to the user")
+	Available = EnumValue(1, "The storage medium of all event subjects "
+		"must be immediately available to the user")
+	Any = EnumValue(2, "The event subjects may or may not be available")
 
-
-class ResultType(object):
+class ResultType(_Enumeration):
 	"""
 	An enumeration class used to define how query results should be returned
 	from the Zeitgeist engine.
 	"""
-	__metaclass__ = EnumMeta
 	
-	MostRecentEvents = enum_factory("All events with the most recent events first")
-	LeastRecentEvents = enum_factory("All events with the oldest ones first")
-	MostRecentSubjects = enum_factory(("One event for each subject only, "
-		"ordered with the most recent events first"))
-	LeastRecentSubjects = enum_factory(("One event for each subject only, "
-		"ordered with oldest events first"))
-	MostPopularSubjects = enum_factory(("One event for each subject only, "
-		"ordered by the popularity of the subject"))
-	LeastPopularSubjects = enum_factory(("One event for each subject only, "
-		"ordered ascendingly by popularity of the subject"))
-	MostPopularActor = enum_factory(("The last event of each different actor,"
-		"ordered by the popularity of the actor"))
-	LeastPopularActor = enum_factory(("The last event of each different actor,"
-		"ordered ascendingly by the popularity of the actor"))
-	MostRecentActor = enum_factory(("The Actor that has been used to most recently"))
-	LeastRecentActor = enum_factory(("The Actor that has been used to least recently"))	
-	MostRecentOrigin = enum_factory(("The last event of each different origin"))
-	LeastRecentOrigin = enum_factory(("The first event of each different origin"))
-	MostPopularOrigin = enum_factory(("The last event of each different origin,"
-		"ordered by the popularity of the origins"))
-	LeastPopularOrigin = enum_factory(("The last event of each different origin,"
-		"ordered ascendingly by the popularity of the origin"))
-	OldestActor = enum_factory(("The first event of each different actor"))
-	MostRecentSubjectInterpretation = enum_factory(("One event for each subject interpretation only, "
-		"ordered with the most recent events first"))
-	LeastRecentSubjectInterpretation = enum_factory(("One event for each subject interpretation only, "
-		"ordered with the least recent events first"))
-	MostPopularSubjectInterpretation = enum_factory(("One event for each subject interpretation only, "
-		"ordered by the popularity of the subject interpretation"))
-	LeastPopularSubjectInterpretation = enum_factory(("One event for each subject interpretation only, "
-		"ordered ascendingly by popularity of the subject interpretation"))
-	MostRecentMimeType = enum_factory(("One event for each mimetype only, "
-		"ordered with the most recent events first"))
-	LeastRecentMimeType = enum_factory(("One event for each mimetype only, "
-		"ordered with the least recent events first"))
-	MostPopularMimeType = enum_factory(("One event for each mimetype only, "
-		"ordered by the popularity of the mimetype"))
-	LeastPopularMimeType = enum_factory(("One event for each mimetype only, "
-		"ordered ascendingly by popularity of the mimetype"))
+	MostRecentEvents = EnumValue(0,
+		"All events with the most recent events first")
+	LeastRecentEvents = EnumValue(1, "All events with the oldest ones first")
+	MostRecentSubjects = EnumValue(2, "One event for each subject only, "
+		"ordered with the most recent events first")
+	LeastRecentSubjects = EnumValue(3, "One event for each subject only, "
+		"ordered with oldest events first")
+	MostPopularSubjects = EnumValue(4, "One event for each subject only, "
+		"ordered by the popularity of the subject")
+	LeastPopularSubjects = EnumValue(5, "One event for each subject only, "
+		"ordered ascendingly by popularity of the subject")
+	MostPopularActor = EnumValue(6, "The last event of each different actor,"
+		"ordered by the popularity of the actor")
+	LeastPopularActor = EnumValue(7, "The last event of each different actor,"
+		"ordered ascendingly by the popularity of the actor")
+	MostRecentActor = EnumValue(8,
+		"The Actor that has been used to most recently")
+	LeastRecentActor = EnumValue(9,
+		"The Actor that has been used to least recently")
+	MostRecentOrigin = EnumValue(10,
+		"The last event of each different subject origin")
+	LeastRecentOrigin = EnumValue(11, "The last event of each different "
+		"subject origin, ordered by least recently used first")
+	MostPopularOrigin = EnumValue(12, "The last event of each different "
+		"subject origin, ordered by the popularity of the origins")
+	LeastPopularOrigin = EnumValue(13, "The last event of each different "
+		"subject origin, ordered ascendingly by the popularity of the origin")
+	OldestActor = EnumValue(14, "The first event of each different actor")
+	MostRecentSubjectInterpretation = EnumValue(15, "One event for each "
+		"subject interpretation only, ordered with the most recent "
+		"events first")
+	LeastRecentSubjectInterpretation = EnumValue(16, "One event for each "
+		"subject interpretation only, ordered with the least recent "
+		"events first")
+	MostPopularSubjectInterpretation = EnumValue(17, "One event for each "
+		"subject interpretation only, ordered by the popularity of the "
+		"subject interpretation")
+	LeastPopularSubjectInterpretation = EnumValue(18, "One event for each "
+		"subject interpretation only, ordered ascendingly by popularity of "
+		"the subject interpretation")
+	MostRecentMimeType = EnumValue(19, "One event for each mimetype only, "
+		"ordered with the most recent events first")
+	LeastRecentMimeType = EnumValue(20, "One event for each mimetype only, "
+		"ordered with the least recent events first")
+	MostPopularMimeType = EnumValue(21, "One event for each mimetype only, "
+		"ordered by the popularity of the mimetype")
+	LeastPopularMimeType = EnumValue(22, "One event for each mimetype only, "
+		"ordered ascendingly by popularity of the mimetype")
+	MostRecentCurrentUri = EnumValue(23, "One event for each subject only "
+		"(by current_uri instead of uri), "
+		"ordered with the most recent events first")
+	LeastRecentCurrentUri = EnumValue(24, "One event for each subject only "
+		"(by current_uri instead of uri), "
+		"ordered with oldest events first")
+	MostPopularCurrentUri = EnumValue(25, "One event for each subject only "
+		"(by current_uri instead of uri), "
+		"ordered by the popularity of the subject")
+	LeastPopularCurrentUri = EnumValue(26, "One event for each subject only "
+		"(by current_uri instead of uri), "
+		"ordered ascendingly by popularity of the subject")
+	MostRecentEventOrigin = EnumValue(27,
+		"The last event of each different origin")
+	LeastRecentEventOrigin = EnumValue(28, "The last event of each "
+		" different origin, ordered by least recently used first")
+	MostPopularEventOrigin = EnumValue(29, "The last event of each "
+		"different origin, ordered by the popularity of the origins")
+	LeastPopularEventOrigin = EnumValue(30, "The last event of each "
+		"different origin, ordered ascendingly by the popularity of the origin")
 
+	# We should eventually migrate over to those names to disambiguate
+	# subject origin and event origin:
+	MostRecentSubjectOrigin = MostRecentOrigin
+	LeastRecentSubjectOrigin = LeastRecentOrigin
+	MostPopularSubjectOrigin = MostPopularOrigin
+	LeastPopularSubjectOrigin = LeastPopularOrigin
 
 INTERPRETATION_DOC = \
 """In general terms the *interpretation* of an event or subject is an abstract
@@ -1148,43 +1210,3 @@ if __name__ == "__main__":
 	print "Success"
 	end_symbols = time.time()
 	print >> sys.stderr, "Import time: %s" % (end_symbols - start_symbols)
-	#~ x = len(Interpretation.get_all_children())
-	#~ y = len(Manifestation.get_all_children())
-	#~ print >> sys.stderr, \
-		#~ ("Overall number of symbols: %i (man.: %i, int.: %i)" %(x+y, y, x))
-	#~ print >> sys.stderr, ("Resolved %i symbols, needed %i iterations" %(initial_count, initial_count-c))
-	#~ print >> sys.stderr, ("Loading symbols took %.4f seconds" %(end_symbols - start_symbols))
-	#~ #
-	#~ # shortcuts
-	#~ EventManifestation = Manifestation.EventManifestation
-	#~ EventInterpretation = Interpretation.EventInterpretation
-	#~ 
-	#~ DataContainer = Interpretation.DataContainer
-	#~ 
-	#~ # testing
-	#~ print dir(EventManifestation)
-	#~ print dir(Manifestation)
-	#~ print EventManifestation.UserActivity
-	#~ 
-	#~ print DataContainer
-	#~ print DataContainer.Filesystem
-	#~ print DataContainer.Filesystem.__doc__
-	#~ 
-	#~ print " OR ".join(DataContainer.get_all_children())
-	#~ print " OR ".join(DataContainer.Filesystem.get_all_children())
-	#~ 
-	#~ print DataContainer.Boo
-	#~ 
-	#~ #Symbol("BOO", DataContainer) #must fail with ValueError
-	#~ #Symbol("Boo", DataContainer) #must fail with ValueError
-	#~ Symbol("Foo", set([DataContainer,]))
-	#~ print DataContainer.Foo
-	#~ 
-	#~ #DataContainer._add_child("booo") #must fail with TypeError
-	#~ 
-	#~ print Interpretation
-	#~ #print Interpretation.get_all_children()
-	#~ import pprint
-	#~ pprint.pprint(Interpretation.Software.get_all_children())
-	#~ 
-	#~ print Interpretation["http://www.semanticdesktop.org/ontologies/2007/03/22/nfo#MindMap"]
